@@ -43,9 +43,17 @@ scm_i_frame_print (SCM frame, SCM port, scm_print_state *pstate)
 {
   scm_puts_unlocked ("#<frame ", port);
   scm_uintprint (SCM_UNPACK (frame), 16, port);
-  scm_putc_unlocked (' ', port);
-  scm_write (scm_frame_procedure (frame), port);
-  /* don't write args, they can get us into trouble. */
+  if (scm_module_system_booted_p)
+    {
+      SCM name = scm_frame_procedure_name (frame);
+
+      if (scm_is_true (name))
+        {
+          scm_putc_unlocked (' ', port);
+          scm_write (name, port);
+        }
+    }
+  /* Don't write args, they can be ridiculously long. */
   scm_puts_unlocked (">", port);
 }
 
@@ -136,16 +144,26 @@ scm_c_frame_closure (enum scm_vm_frame_kind kind, const struct scm_frame *frame)
   return SCM_BOOL_F;
 }
 
-SCM_DEFINE (scm_frame_procedure, "frame-procedure", 1, 0, 0,
+static SCM frame_procedure_name_var;
+
+static void
+init_frame_procedure_name_var (void)
+{
+  frame_procedure_name_var
+    = scm_c_private_lookup ("system vm frame", "frame-procedure-name");
+}
+
+SCM_DEFINE (scm_frame_procedure_name, "frame-procedure-name", 1, 0, 0,
 	    (SCM frame),
 	    "")
-#define FUNC_NAME s_scm_frame_procedure
+#define FUNC_NAME s_scm_frame_procedure_name
 {
+  static scm_i_pthread_once_t once = SCM_I_PTHREAD_ONCE_INIT;
+  scm_i_pthread_once (&once, init_frame_procedure_name_var);
+
   SCM_VALIDATE_VM_FRAME (1, frame);
 
-  /* FIXME: Retrieve procedure from address?  */
-  return scm_c_frame_closure (SCM_VM_FRAME_KIND (frame),
-                              SCM_VM_FRAME_DATA (frame));
+  return scm_call_1 (scm_variable_ref (frame_procedure_name_var), frame);
 }
 #undef FUNC_NAME
 
@@ -204,9 +222,9 @@ SCM_DEFINE (scm_frame_source, "frame-source", 1, 0, 0,
 }
 #undef FUNC_NAME
 
-SCM_DEFINE (scm_frame_num_locals, "frame-num-locals", 1, 0, 0,
-	    (SCM frame),
-	    "")
+static const char s_scm_frame_num_locals[] = "frame-num-locals";
+static SCM
+scm_frame_num_locals (SCM frame)
 #define FUNC_NAME s_scm_frame_num_locals
 {
   union scm_vm_stack_element *fp, *sp;
@@ -220,45 +238,104 @@ SCM_DEFINE (scm_frame_num_locals, "frame-num-locals", 1, 0, 0,
 }
 #undef FUNC_NAME
 
-SCM_DEFINE (scm_frame_local_ref, "frame-local-ref", 2, 0, 0,
-	    (SCM frame, SCM index),
-	    "")
+enum stack_item_representation
+  {
+    STACK_ITEM_SCM = 0,
+    STACK_ITEM_F64 = 1,
+    STACK_ITEM_U64 = 2,
+    STACK_ITEM_S64 = 3
+  };
+
+static enum stack_item_representation
+scm_to_stack_item_representation (SCM x, const char *subr, int pos)
+{
+  if (scm_is_eq (x, scm_from_latin1_symbol ("scm")))
+    return STACK_ITEM_SCM;
+  if (scm_is_eq (x, scm_from_latin1_symbol ("f64")))
+    return STACK_ITEM_F64;
+  if (scm_is_eq (x, scm_from_latin1_symbol ("u64")))
+    return STACK_ITEM_U64;
+  if (scm_is_eq (x, scm_from_latin1_symbol ("s64")))
+    return STACK_ITEM_S64;
+
+  scm_wrong_type_arg (subr, pos, x);
+  return 0;  /* Not reached.  */
+}
+
+static const char s_scm_frame_local_ref[] = "frame-local-ref";
+static SCM
+scm_frame_local_ref (SCM frame, SCM index, SCM representation)
 #define FUNC_NAME s_scm_frame_local_ref
 {
   union scm_vm_stack_element *fp, *sp;
   unsigned int i;
+  enum stack_item_representation repr;
 
   SCM_VALIDATE_VM_FRAME (1, frame);
   SCM_VALIDATE_UINT_COPY (2, index, i);
-
-  fp = SCM_VM_FRAME_FP (frame);
-  sp = SCM_VM_FRAME_SP (frame);
-
-  if (i < SCM_FRAME_NUM_LOCALS (fp, sp))
-    return SCM_FRAME_LOCAL (fp, i);
-
-  SCM_OUT_OF_RANGE (SCM_ARG2, index);
-}
-#undef FUNC_NAME
-
-/* Need same not-yet-active frame logic here as in frame-num-locals */
-SCM_DEFINE (scm_frame_local_set_x, "frame-local-set!", 3, 0, 0,
-	    (SCM frame, SCM index, SCM val),
-	    "")
-#define FUNC_NAME s_scm_frame_local_set_x
-{
-  union scm_vm_stack_element *fp, *sp;
-  unsigned int i;
-
-  SCM_VALIDATE_VM_FRAME (1, frame);
-  SCM_VALIDATE_UINT_COPY (2, index, i);
+  repr = scm_to_stack_item_representation (representation, FUNC_NAME, SCM_ARG3);
 
   fp = SCM_VM_FRAME_FP (frame);
   sp = SCM_VM_FRAME_SP (frame);
 
   if (i < SCM_FRAME_NUM_LOCALS (fp, sp))
     {
-      SCM_FRAME_LOCAL (fp, i) = val;
+      union scm_vm_stack_element *item = SCM_FRAME_SLOT (fp, i);
+      switch (repr)
+        {
+          case STACK_ITEM_SCM:
+            return item->as_scm;
+          case STACK_ITEM_F64:
+            return scm_from_double (item->as_f64);
+          case STACK_ITEM_U64:
+            return scm_from_uint64 (item->as_u64);
+          case STACK_ITEM_S64:
+            return scm_from_int64 (item->as_s64);
+          default:
+            abort();
+        }
+    }
+
+  SCM_OUT_OF_RANGE (SCM_ARG2, index);
+}
+#undef FUNC_NAME
+
+static const char s_scm_frame_local_set_x[] = "frame-local-set!";
+static SCM
+scm_frame_local_set_x (SCM frame, SCM index, SCM val, SCM representation)
+#define FUNC_NAME s_scm_frame_local_set_x
+{
+  union scm_vm_stack_element *fp, *sp;
+  unsigned int i;
+  enum stack_item_representation repr;
+
+  SCM_VALIDATE_VM_FRAME (1, frame);
+  SCM_VALIDATE_UINT_COPY (2, index, i);
+  repr = scm_to_stack_item_representation (representation, FUNC_NAME, SCM_ARG3);
+
+  fp = SCM_VM_FRAME_FP (frame);
+  sp = SCM_VM_FRAME_SP (frame);
+
+  if (i < SCM_FRAME_NUM_LOCALS (fp, sp))
+    {
+      union scm_vm_stack_element *item = SCM_FRAME_SLOT (fp, i);
+      switch (repr)
+        {
+          case STACK_ITEM_SCM:
+            item->as_scm = val;
+            break;
+          case STACK_ITEM_F64:
+            item->as_f64 = scm_to_double (val);
+            break;
+          case STACK_ITEM_U64:
+            item->as_u64 = scm_to_uint64 (val);
+            break;
+          case STACK_ITEM_S64:
+            item->as_s64 = scm_to_int64 (val);
+            break;
+          default:
+            abort();
+        }
       return SCM_UNSPECIFIED;
     }
 
@@ -344,11 +421,8 @@ scm_c_frame_previous (enum scm_vm_frame_kind kind, struct scm_frame *frame)
   frame->sp_offset = stack_top - new_sp;
   frame->ip = SCM_FRAME_RETURN_ADDRESS (this_fp);
 
-  {
-    SCM proc = scm_c_frame_closure (kind, frame);
-    if (SCM_PROGRAM_P (proc) && SCM_PROGRAM_IS_BOOT (proc))
-      goto again;
-  }
+  if (scm_i_vm_is_boot_continuation_code (frame->ip))
+    goto again;
 
   return 1;
 }
@@ -374,12 +448,28 @@ SCM_DEFINE (scm_frame_previous, "frame-previous", 1, 0, 0,
 #undef FUNC_NAME
 
 
+static void
+scm_init_frames_builtins (void *unused)
+{
+  scm_c_define_gsubr (s_scm_frame_num_locals, 1, 0, 0,
+                      (scm_t_subr) scm_frame_num_locals);
+  scm_c_define_gsubr (s_scm_frame_local_ref, 3, 0, 0,
+                      (scm_t_subr) scm_frame_local_ref);
+  scm_c_define_gsubr (s_scm_frame_local_set_x, 4, 0, 0,
+                      (scm_t_subr) scm_frame_local_set_x);
+}
+
 void
 scm_init_frames (void)
 {
 #ifndef SCM_MAGIC_SNARFER
 #include "libguile/frames.x"
 #endif
+
+  scm_c_register_extension ("libguile-" SCM_EFFECTIVE_VERSION,
+                            "scm_init_frames_builtins",
+                            scm_init_frames_builtins,
+                            NULL);
 }
 
 /*
