@@ -81,10 +81,6 @@
 #include "libguile/threads.h"
 
 
-#ifdef __MINGW32__
-# include "posix-w32.h"
-#endif
-
 #if HAVE_SYS_WAIT_H
 # include <sys/wait.h>
 #endif
@@ -630,7 +626,6 @@ SCM_DEFINE (scm_setrlimit, "setrlimit", 3, 0, 0,
 #endif /* HAVE_GETRLIMIT */
 
 
-#ifdef HAVE_KILL
 SCM_DEFINE (scm_kill, "kill", 2, 0, 0,
             (SCM pid, SCM sig),
 	    "Sends a signal to the specified process or group of processes.\n\n"
@@ -658,12 +653,30 @@ SCM_DEFINE (scm_kill, "kill", 2, 0, 0,
 #define FUNC_NAME s_scm_kill
 {
   /* Signal values are interned in scm_init_posix().  */
+#ifdef HAVE_KILL
   if (kill (scm_to_int (pid), scm_to_int  (sig)) != 0)
     SCM_SYSERROR;
+#else
+  /* Mingw has raise(), but not kill().  (Other raw DOS environments might
+     be similar.)  Use raise() when the requested pid is our own process,
+     otherwise bomb.  */
+  if (scm_to_int (pid) == getpid ())
+    {
+      if (raise (scm_to_int (sig)) != 0)
+        {
+        err:
+          SCM_SYSERROR;
+        }
+      else
+        {
+          errno = ENOSYS;
+          goto err;
+        }
+    }
+#endif
   return SCM_UNSPECIFIED;
 }
 #undef FUNC_NAME
-#endif
 
 #ifdef HAVE_WAITPID
 SCM_DEFINE (scm_waitpid, "waitpid", 1, 1, 0,
@@ -722,7 +735,7 @@ SCM_DEFINE (scm_waitpid, "waitpid", 1, 1, 0,
 #undef FUNC_NAME
 #endif /* HAVE_WAITPID */
 
-#ifdef WIFEXITED
+#ifndef __MINGW32__
 SCM_DEFINE (scm_status_exit_val, "status:exit-val", 1, 0, 0, 
             (SCM status),
 	    "Return the exit status value, as would be set if a process\n"
@@ -741,9 +754,7 @@ SCM_DEFINE (scm_status_exit_val, "status:exit-val", 1, 0, 0,
     return SCM_BOOL_F;
 }
 #undef FUNC_NAME
-#endif /* WIFEXITED */
 
-#ifdef WIFSIGNALED
 SCM_DEFINE (scm_status_term_sig, "status:term-sig", 1, 0, 0, 
             (SCM status),
 	    "Return the signal number which terminated the process, if any,\n"
@@ -759,9 +770,7 @@ SCM_DEFINE (scm_status_term_sig, "status:term-sig", 1, 0, 0,
     return SCM_BOOL_F;
 }
 #undef FUNC_NAME
-#endif /* WIFSIGNALED */
 
-#ifdef WIFSTOPPED
 SCM_DEFINE (scm_status_stop_sig, "status:stop-sig", 1, 0, 0, 
             (SCM status),
 	    "Return the signal number which stopped the process, if any,\n"
@@ -777,7 +786,7 @@ SCM_DEFINE (scm_status_stop_sig, "status:stop-sig", 1, 0, 0,
     return SCM_BOOL_F;
 }
 #undef FUNC_NAME
-#endif /* WIFSTOPPED */
+#endif /* __MINGW32__ */
 
 #ifdef HAVE_GETPPID
 SCM_DEFINE (scm_getppid, "getppid", 0, 0, 0,
@@ -791,7 +800,8 @@ SCM_DEFINE (scm_getppid, "getppid", 0, 0, 0,
 #undef FUNC_NAME
 #endif /* HAVE_GETPPID */
 
-#ifdef HAVE_GETUID
+
+#ifndef __MINGW32__
 SCM_DEFINE (scm_getuid, "getuid", 0, 0, 0,
             (),
 	    "Return an integer representing the current real user ID.")
@@ -800,9 +810,9 @@ SCM_DEFINE (scm_getuid, "getuid", 0, 0, 0,
   return scm_from_int (getuid ());
 }
 #undef FUNC_NAME
-#endif /* HAVE_GETUID */
 
-#ifdef HAVE_GETGID
+
+
 SCM_DEFINE (scm_getgid, "getgid", 0, 0, 0,
             (),
 	    "Return an integer representing the current real group ID.")
@@ -811,9 +821,9 @@ SCM_DEFINE (scm_getgid, "getgid", 0, 0, 0,
   return scm_from_int (getgid ());
 }
 #undef FUNC_NAME
-#endif /* HAVE_GETGID */
 
-#ifdef HAVE_GETUID
+
+
 SCM_DEFINE (scm_geteuid, "geteuid", 0, 0, 0,
             (),
 	    "Return an integer representing the current effective user ID.\n"
@@ -829,9 +839,8 @@ SCM_DEFINE (scm_geteuid, "geteuid", 0, 0, 0,
 #endif
 }
 #undef FUNC_NAME
-#endif /* HAVE_GETUID */
 
-#ifdef HAVE_GETGID
+
 SCM_DEFINE (scm_getegid, "getegid", 0, 0, 0,
             (),
 	    "Return an integer representing the current effective group ID.\n"
@@ -847,9 +856,8 @@ SCM_DEFINE (scm_getegid, "getegid", 0, 0, 0,
 #endif
 }
 #undef FUNC_NAME
-#endif /* HAVE_GETGID */
 
-#ifdef HAVE_SETUID
+
 SCM_DEFINE (scm_setuid, "setuid", 1, 0, 0, 
             (SCM id),
 	    "Sets both the real and effective user IDs to the integer @var{id}, provided\n"
@@ -862,9 +870,7 @@ SCM_DEFINE (scm_setuid, "setuid", 1, 0, 0,
   return SCM_UNSPECIFIED;
 }
 #undef FUNC_NAME
-#endif /* HAVE_SETUID */
 
-#ifdef HAVE_SETGID
 SCM_DEFINE (scm_setgid, "setgid", 1, 0, 0, 
             (SCM id),
 	    "Sets both the real and effective group IDs to the integer @var{id}, provided\n"
@@ -877,9 +883,7 @@ SCM_DEFINE (scm_setgid, "setgid", 1, 0, 0,
   return SCM_UNSPECIFIED;
 }
 #undef FUNC_NAME
-#endif /* HAVE_SETGID */
 
-#ifdef HAVE_SETUID
 SCM_DEFINE (scm_seteuid, "seteuid", 1, 0, 0, 
             (SCM id),
 	    "Sets the effective user ID to the integer @var{id}, provided the process\n"
@@ -901,9 +905,10 @@ SCM_DEFINE (scm_seteuid, "seteuid", 1, 0, 0,
   return SCM_UNSPECIFIED;
 }
 #undef FUNC_NAME
-#endif /* HAVE_SETUID */
+#endif /* __MINGW32__ */
 
-#ifdef HAVE_SETGID
+
+#ifdef HAVE_SETEGID
 SCM_DEFINE (scm_setegid, "setegid", 1, 0, 0,
             (SCM id),
 	    "Sets the effective group ID to the integer @var{id}, provided the process\n"
@@ -926,7 +931,8 @@ SCM_DEFINE (scm_setegid, "setegid", 1, 0, 0,
     
 }
 #undef FUNC_NAME
-#endif /* HAVE_SETGID */
+#endif
+
 
 #ifdef HAVE_GETPGRP
 SCM_DEFINE (scm_getpgrp, "getpgrp", 0, 0, 0,
@@ -942,6 +948,7 @@ SCM_DEFINE (scm_getpgrp, "getpgrp", 0, 0, 0,
 #undef FUNC_NAME
 #endif /* HAVE_GETPGRP */
 
+
 #ifdef HAVE_SETPGID
 SCM_DEFINE (scm_setpgid, "setpgid", 2, 0, 0, 
             (SCM pid, SCM pgid),
@@ -1217,6 +1224,7 @@ SCM_DEFINE (scm_fork, "primitive-fork", 0, 0, 0,
 #define FUNC_NAME s_scm_fork
 {
   int pid;
+  scm_i_finalizer_pre_fork ();
   if (scm_ilength (scm_all_threads ()) != 1)
     /* Other threads may be holding on to resources that Guile needs --
        it is not safe to permit one thread to fork while others are
@@ -1239,96 +1247,10 @@ SCM_DEFINE (scm_fork, "primitive-fork", 0, 0, 0,
   return scm_from_int (pid);
 }
 #undef FUNC_NAME
-#endif /* HAVE_FORK */
 
-#ifdef HAVE_FORK
-#define HAVE_START_CHILD 1
 /* Since Guile uses threads, we have to be very careful to avoid calling
    functions that are not async-signal-safe in the child.  That's why
    this function is implemented in C.  */
-static pid_t
-start_child (const char *exec_file, char **exec_argv,
-	     int reading, int c2p[2], int writing, int p2c[2],
-             int in, int out, int err)
-{
-  int pid;
-  int max_fd = 1024;
-
-#if defined (HAVE_GETRLIMIT) && defined (RLIMIT_NOFILE)
-  {
-    struct rlimit lim = { 0, 0 };
-    if (getrlimit (RLIMIT_NOFILE, &lim) == 0)
-      max_fd = lim.rlim_cur;
-  }
-#endif
-
-  pid = fork ();
-
-  if (pid != 0)
-    /* The parent, with either and error (pid == -1), or the PID of the
-       child.  Return directly in either case.  */
-    return pid;
-
-  /* The child.  */
-  if (reading)
-    close (c2p[0]);
-  if (writing)
-    close (p2c[1]);
-
-  /* Close all file descriptors in ports inherited from the parent
-     except for in, out, and err.  Heavy-handed, but robust.  */
-  while (max_fd--)
-    if (max_fd != in && max_fd != out && max_fd != err)
-      close (max_fd);
-
-  /* Ignore errors on these open() calls.  */
-  if (in == -1)
-    in = open ("/dev/null", O_RDONLY);
-  if (out == -1)
-    out = open ("/dev/null", O_WRONLY);
-  if (err == -1)
-    err = open ("/dev/null", O_WRONLY);
-
-  if (in > 0)
-    {
-      if (out == 0)
-        do out = dup (out); while (errno == EINTR);
-      if (err == 0)
-        do err = dup (err); while (errno == EINTR);
-      do dup2 (in, 0); while (errno == EINTR);
-      close (in);
-    }
-  if (out > 1)
-    {
-      if (err == 1)
-        do err = dup (err); while (errno == EINTR);
-      do dup2 (out, 1); while (errno == EINTR);
-      close (out);
-    }
-  if (err > 2)
-    {
-      do dup2 (err, 2); while (errno == EINTR);
-      close (err);
-    }
-
-  execvp (exec_file, exec_argv);
-
-  /* The exec failed!  There is nothing sensible to do.  */
-  if (err > 0)
-    {
-      char *msg = strerror (errno);
-      fprintf (fdopen (err, "a"), "In execvp of %s: %s\n",
-               exec_file, msg);
-    }
-
-  _exit (EXIT_FAILURE);
-
-  /* Not reached.  */
-  return -1;
-}
-#endif
-
-#ifdef HAVE_START_CHILD
 static SCM
 scm_open_process (SCM mode, SCM prog, SCM args)
 #define FUNC_NAME "open-process"
@@ -1341,7 +1263,7 @@ scm_open_process (SCM mode, SCM prog, SCM args)
   int pid;
   char *exec_file;
   char **exec_argv;
-  SCM read_port = SCM_BOOL_F, write_port = SCM_BOOL_F;
+  int max_fd = 1024;
 
   exec_file = scm_to_locale_string (prog);
   exec_argv = scm_i_allocate_string_pointers (scm_cons (prog, args));
@@ -1390,8 +1312,15 @@ scm_open_process (SCM mode, SCM prog, SCM args)
       in = SCM_FPORT_FDES (port);
   }
 
-  pid = start_child (exec_file, exec_argv, reading, c2p, writing, p2c,
-                     in, out, err);
+#if defined (HAVE_GETRLIMIT) && defined (RLIMIT_NOFILE)
+  {
+    struct rlimit lim = { 0, 0 };
+    if (getrlimit (RLIMIT_NOFILE, &lim) == 0)
+      max_fd = lim.rlim_cur;
+  }
+#endif
+
+  pid = fork ();
 
   if (pid == -1)
     {
@@ -1411,26 +1340,91 @@ scm_open_process (SCM mode, SCM prog, SCM args)
       SCM_SYSERROR;
     }
 
-  /* There is no sense in catching errors on close().  */
-  if (reading)
+  if (pid)
+    /* Parent. */
     {
-      close (c2p[1]);
-      read_port = scm_fdes_to_port (c2p[0], "r0", sym_read_pipe);
-    }
-  if (writing)
-    {
-      close (p2c[0]);
-      write_port = scm_fdes_to_port (p2c[1], "w0", sym_write_pipe);
+      SCM read_port = SCM_BOOL_F, write_port = SCM_BOOL_F;
+
+      /* There is no sense in catching errors on close().  */
+      if (reading)
+        {
+          close (c2p[1]);
+          read_port = scm_fdes_to_port (c2p[0], "r0", sym_read_pipe);
+        }
+      if (writing)
+        {
+          close (p2c[0]);
+          write_port = scm_fdes_to_port (p2c[1], "w0", sym_write_pipe);
+        }
+
+      return scm_values
+        (scm_list_3 (read_port, write_port, scm_from_int (pid)));
     }
 
-  return scm_values (scm_list_3 (read_port,
-                                 write_port,
-                                 scm_from_int (pid)));
+  /* The child.  */
+  if (reading)
+    close (c2p[0]);
+  if (writing)
+    close (p2c[1]);
+
+  /* Close all file descriptors in ports inherited from the parent
+     except for in, out, and err.  Heavy-handed, but robust.  */
+  while (max_fd--)
+    if (max_fd != in && max_fd != out && max_fd != err)
+      close (max_fd);
+
+  /* Ignore errors on these open() calls.  */
+  if (in == -1)
+    in = open ("/dev/null", O_RDONLY);
+  if (out == -1)
+    out = open ("/dev/null", O_WRONLY);
+  if (err == -1)
+    err = open ("/dev/null", O_WRONLY);
+    
+  if (in > 0)
+    {
+      if (out == 0)
+        do out = dup (out); while (errno == EINTR);
+      if (err == 0)
+        do err = dup (err); while (errno == EINTR);
+      do dup2 (in, 0); while (errno == EINTR);
+      close (in);
+    }
+  if (out > 1)
+    {
+      if (err == 1)
+        do err = dup (err); while (errno == EINTR);
+      do dup2 (out, 1); while (errno == EINTR);
+      close (out);
+    }
+  if (err > 2)
+    {
+      do dup2 (err, 2); while (errno == EINTR);
+      close (err);
+    }
+
+  execvp (exec_file, exec_argv);
+
+  /* The exec failed!  There is nothing sensible to do.  */
+  if (err > 0)
+    {
+      char *msg = strerror (errno);
+      fprintf (fdopen (err, "a"), "In execlp of %s: %s\n",
+               exec_file, msg);
+    }
+
+  _exit (EXIT_FAILURE);
+  /* Not reached.  */
+  return SCM_BOOL_F;
 }
 #undef FUNC_NAME
-#endif /* HAVE_START_CHILD */
+#endif /* HAVE_FORK */
 
-#ifdef HAVE_UNAME
+#ifdef __MINGW32__
+# include "win32-uname.h"
+#endif
+
+#if defined (HAVE_UNAME) || defined (__MINGW32__)
 SCM_DEFINE (scm_uname, "uname", 0, 0, 0,
             (),
 	    "Return an object with some information about the computer\n"
@@ -1979,6 +1973,7 @@ SCM_DEFINE (scm_setpriority, "setpriority", 3, 0, 0,
 #endif /* HAVE_SETPRIORITY */
 
 #ifdef HAVE_SCHED_GETAFFINITY
+
 static SCM
 cpu_set_to_bitvector (const cpu_set_t *cs)
 {
@@ -2003,7 +1998,10 @@ SCM_DEFINE (scm_getaffinity, "getaffinity", 1, 0, 0,
 	    "process @var{pid}.  Each CPU the process has affinity with\n"
 	    "has its corresponding bit set in the returned bitvector.\n"
 	    "The number of bits set is a good estimate of how many CPUs\n"
-	    "Guile can use without stepping on other processes' toes.")
+	    "Guile can use without stepping on other processes' toes.\n\n"
+	    "Currently this procedure is only defined on GNU variants\n"
+	    "(@pxref{CPU Affinity, @code{sched_getaffinity},, libc, The\n"
+	    "GNU C Library Reference Manual}).\n")
 #define FUNC_NAME s_scm_getaffinity
 {
   int err;
@@ -2017,14 +2015,19 @@ SCM_DEFINE (scm_getaffinity, "getaffinity", 1, 0, 0,
   return cpu_set_to_bitvector (&cs);
 }
 #undef FUNC_NAME
+
 #endif /* HAVE_SCHED_GETAFFINITY */
 
 #ifdef HAVE_SCHED_SETAFFINITY
+
 SCM_DEFINE (scm_setaffinity, "setaffinity", 2, 0, 0,
 	    (SCM pid, SCM mask),
 	    "Install the CPU affinity mask @var{mask}, a bitvector, for\n"
 	    "the process or thread with ID @var{pid}.  The return value\n"
-	    "is unspecified.")
+	    "is unspecified.\n\n"
+	    "Currently this procedure is only defined on GNU variants\n"
+	    "(@pxref{CPU Affinity, @code{sched_setaffinity},, libc, The\n"
+	    "GNU C Library Reference Manual}).\n")
 #define FUNC_NAME s_scm_setaffinity
 {
   cpu_set_t cs;
@@ -2053,6 +2056,7 @@ SCM_DEFINE (scm_setaffinity, "setaffinity", 2, 0, 0,
   return SCM_UNSPECIFIED;
 }
 #undef FUNC_NAME
+
 #endif /* HAVE_SCHED_SETAFFINITY */
 
 
@@ -2232,13 +2236,13 @@ SCM_DEFINE (scm_gethostname, "gethostname", 0, 0, 0,
 #endif /* HAVE_GETHOSTNAME */
 
 
-#ifdef HAVE_START_CHILD
+#ifdef HAVE_FORK
 static void
 scm_init_popen (void)
 {
   scm_c_define_gsubr ("open-process", 2, 0, 1, scm_open_process);
 }
-#endif /* HAVE_START_CHILD */
+#endif
 
 void
 scm_init_posix ()
@@ -2337,14 +2341,11 @@ scm_init_posix ()
 
 #ifdef HAVE_FORK
   scm_add_feature ("fork");
-#endif	/* HAVE_FORK */
-#ifdef HAVE_START_CHILD
-  scm_add_feature ("popen");
   scm_c_register_extension ("libguile-" SCM_EFFECTIVE_VERSION,
                             "scm_init_popen",
 			    (scm_t_extension_init_func) scm_init_popen,
 			    NULL);
-#endif /* HAVE_START_CHILD */
+#endif	/* HAVE_FORK */
 }
 
 /*
