@@ -443,10 +443,13 @@ static void vm_error_wrong_type_apply (SCM proc) SCM_NORETURN SCM_NOINLINE;
 static void vm_error_stack_underflow (void) SCM_NORETURN SCM_NOINLINE;
 static void vm_error_improper_list (SCM x) SCM_NORETURN SCM_NOINLINE;
 static void vm_error_not_a_pair (const char *subr, SCM x) SCM_NORETURN SCM_NOINLINE;
+static void vm_error_not_a_string (const char *subr, SCM x) SCM_NORETURN SCM_NOINLINE;
 static void vm_error_not_a_bytevector (const char *subr, SCM x) SCM_NORETURN SCM_NOINLINE;
 static void vm_error_not_a_struct (const char *subr, SCM x) SCM_NORETURN SCM_NOINLINE;
 static void vm_error_not_a_vector (const char *subr, SCM v) SCM_NORETURN SCM_NOINLINE;
 static void vm_error_out_of_range (const char *subr, SCM k) SCM_NORETURN SCM_NOINLINE;
+static void vm_error_out_of_range_uint64 (const char *subr, scm_t_uint64 idx) SCM_NORETURN SCM_NOINLINE;
+static void vm_error_out_of_range_int64 (const char *subr, scm_t_int64 idx) SCM_NORETURN SCM_NOINLINE;
 static void vm_error_no_values (void) SCM_NORETURN SCM_NOINLINE;
 static void vm_error_not_enough_values (void) SCM_NORETURN SCM_NOINLINE;
 static void vm_error_wrong_number_of_values (scm_t_uint32 expected) SCM_NORETURN SCM_NOINLINE;
@@ -559,6 +562,12 @@ vm_error_not_a_pair (const char *subr, SCM x)
   scm_wrong_type_arg_msg (subr, 1, x, "pair");
 }
 
+static void
+vm_error_not_a_string (const char *subr, SCM x)
+{
+  scm_wrong_type_arg_msg (subr, 1, x, "string");
+}
+
 static void
 vm_error_not_a_bytevector (const char *subr, SCM x)
 {
@@ -584,6 +593,18 @@ vm_error_out_of_range (const char *subr, SCM k)
   scm_out_of_range (subr, k);
 }
 
+static void
+vm_error_out_of_range_uint64 (const char *subr, scm_t_uint64 idx)
+{
+  scm_out_of_range (subr, scm_from_uint64 (idx));
+}
+
+static void
+vm_error_out_of_range_int64 (const char *subr, scm_t_int64 idx)
+{
+  scm_out_of_range (subr, scm_from_int64 (idx));
+}
+
 static void
 vm_error_no_values (void)
 {
@@ -630,6 +651,10 @@ static const scm_t_uint32 vm_boot_continuation_code[] = {
   SCM_PACK_OP_24 (halt, 0)
 };
 
+static const scm_t_uint32 vm_apply_non_program_code[] = {
+  SCM_PACK_OP_24 (apply_non_program, 0)
+};
+
 static const scm_t_uint32 vm_builtin_apply_code[] = {
   SCM_PACK_OP_24 (assert_nargs_ge, 3),
   SCM_PACK_OP_24 (tail_apply, 0), /* proc in r1, args from r2 */
@@ -661,6 +686,12 @@ static const scm_t_uint32 vm_builtin_call_with_current_continuation_code[] = {
 };
 
 
+int
+scm_i_vm_is_boot_continuation_code (scm_t_uint32 *ip)
+{
+  return ip == vm_boot_continuation_code;
+}
+
 static SCM
 scm_vm_builtin_ref (unsigned idx)
 {
@@ -895,31 +926,31 @@ return_unused_stack_to_os (struct scm_vm *vp)
 #endif
 }
 
-#define DEAD_SLOT_MAP_CACHE_SIZE 32U
-struct dead_slot_map_cache_entry
+#define SLOT_MAP_CACHE_SIZE 32U
+struct slot_map_cache_entry
 {
   scm_t_uint32 *ip;
   const scm_t_uint8 *map;
 };
 
-struct dead_slot_map_cache
+struct slot_map_cache
 {
-  struct dead_slot_map_cache_entry entries[DEAD_SLOT_MAP_CACHE_SIZE];
+  struct slot_map_cache_entry entries[SLOT_MAP_CACHE_SIZE];
 };
 
 static const scm_t_uint8 *
-find_dead_slot_map (scm_t_uint32 *ip, struct dead_slot_map_cache *cache)
+find_slot_map (scm_t_uint32 *ip, struct slot_map_cache *cache)
 {
   /* The lower two bits should be zero.  FIXME: Use a better hash
      function; we don't expose scm_raw_hashq currently.  */
-  size_t slot = (((scm_t_uintptr) ip) >> 2) % DEAD_SLOT_MAP_CACHE_SIZE;
+  size_t slot = (((scm_t_uintptr) ip) >> 2) % SLOT_MAP_CACHE_SIZE;
   const scm_t_uint8 *map;
 
   if (cache->entries[slot].ip == ip)
     map = cache->entries[slot].map;
   else
     {
-      map = scm_find_dead_slot_map_unlocked (ip);
+      map = scm_find_slot_map_unlocked (ip);
       cache->entries[slot].ip = ip;
       cache->entries[slot].map = map;
     }
@@ -927,21 +958,29 @@ find_dead_slot_map (scm_t_uint32 *ip, struct dead_slot_map_cache *cache)
   return map;
 }
 
+enum slot_desc
+  {
+    SLOT_DESC_DEAD = 0,
+    SLOT_DESC_LIVE_RAW = 1,
+    SLOT_DESC_LIVE_SCM = 2,
+    SLOT_DESC_UNUSED = 3
+  };
+
 /* Mark the active VM stack region.  */
 struct GC_ms_entry *
 scm_i_vm_mark_stack (struct scm_vm *vp, struct GC_ms_entry *mark_stack_ptr,
                      struct GC_ms_entry *mark_stack_limit)
 {
   union scm_vm_stack_element *sp, *fp;
-  /* The first frame will be marked conservatively (without a dead
-     slot map).  This is because GC can happen at any point within the
-     hottest activation, due to multiple threads or per-instruction
-     hooks, and providing dead slot maps for all points in a program
-     would take a prohibitive amount of space.  */
-  const scm_t_uint8 *dead_slots = NULL;
+  /* The first frame will be marked conservatively (without a slot map).
+     This is because GC can happen at any point within the hottest
+     activation, due to multiple threads or per-instruction hooks, and
+     providing slot maps for all points in a program would take a
+     prohibitive amount of space.  */
+  const scm_t_uint8 *slot_map = NULL;
   void *upper = (void *) GC_greatest_plausible_heap_addr;
   void *lower = (void *) GC_least_plausible_heap_addr;
-  struct dead_slot_map_cache cache;
+  struct slot_map_cache cache;
 
   memset (&cache, 0, sizeof (cache));
 
@@ -953,24 +992,29 @@ scm_i_vm_mark_stack (struct scm_vm *vp, struct GC_ms_entry *mark_stack_ptr,
       size_t slot = nlocals - 1;
       for (slot = nlocals - 1; sp < fp; sp++, slot--)
         {
-          if (SCM_NIMP (sp->as_scm) &&
-              sp->as_ptr >= lower && sp->as_ptr <= upper)
-            {
-              if (dead_slots)
-                {
-                  if (dead_slots[slot / 8U] & (1U << (slot % 8U)))
-                    {
-                      /* This value may become dead as a result of GC,
-                         so we can't just leave it on the stack.  */
-                      sp->as_scm = SCM_UNSPECIFIED;
-                      continue;
-                    }
-                }
+          enum slot_desc desc = SLOT_DESC_LIVE_SCM;
 
-              mark_stack_ptr = GC_mark_and_push (sp->as_ptr,
-                                                 mark_stack_ptr,
-                                                 mark_stack_limit,
-                                                 NULL);
+          if (slot_map)
+            desc = (slot_map[slot / 4U] >> ((slot % 4U) * 2)) & 3U;
+
+          switch (desc)
+            {
+            case SLOT_DESC_LIVE_RAW:
+              break;
+            case SLOT_DESC_UNUSED:
+            case SLOT_DESC_LIVE_SCM:
+              if (SCM_NIMP (sp->as_scm) &&
+                  sp->as_ptr >= lower && sp->as_ptr <= upper)
+                mark_stack_ptr = GC_mark_and_push (sp->as_ptr,
+                                                   mark_stack_ptr,
+                                                   mark_stack_limit,
+                                                   NULL);
+              break;
+            case SLOT_DESC_DEAD:
+              /* This value may become dead as a result of GC,
+                 so we can't just leave it on the stack.  */
+              sp->as_scm = SCM_UNSPECIFIED;
+              break;
             }
         }
       sp = SCM_FRAME_PREVIOUS_SP (fp);
@@ -978,7 +1022,7 @@ scm_i_vm_mark_stack (struct scm_vm *vp, struct GC_ms_entry *mark_stack_ptr,
          Note that there may be other reasons to not have a dead slots
          map, e.g. if all of the frame's slots below the callee frame
          are live.  */
-      dead_slots = find_dead_slot_map (SCM_FRAME_RETURN_ADDRESS (fp), &cache);
+      slot_map = find_slot_map (SCM_FRAME_RETURN_ADDRESS (fp), &cache);
     }
 
   return_unused_stack_to_os (vp);
