@@ -122,25 +122,11 @@
                     ((src . dst) (emit-mov asm (from-sp dst) (from-sp src))))
                    (lookup-parallel-moves label allocation))
          (emit-tail-call-label asm (1+ (length args)) k))
-        (($ $values ())
-         (emit-reset-frame asm 1)
-         (emit-return-values asm))
-        (($ $values (arg))
-         (if (maybe-slot arg)
-             (emit-return asm (from-sp (slot arg)))
-             (begin
-               (when (< frame-size 2)
-                 (emit-alloc-frame asm 2))
-               (emit-load-constant asm (from-sp 1) (constant arg))
-               (emit-return asm (from-sp 1)))))
         (($ $values args)
          (for-each (match-lambda
                     ((src . dst) (emit-mov asm (from-sp dst) (from-sp src))))
                    (lookup-parallel-moves label allocation))
-         (emit-reset-frame asm (1+ (length args)))
-         (emit-return-values asm))
-        (($ $primcall 'return (arg))
-         (emit-return asm (from-sp (slot arg))))))
+         (emit-return-values asm (1+ (length args))))))
 
     (define (compile-value label exp dst)
       (match exp
@@ -154,6 +140,8 @@
          (emit-make-closure asm (from-sp dst) k nfree))
         (($ $primcall 'current-module)
          (emit-current-module asm (from-sp dst)))
+        (($ $primcall 'current-thread)
+         (emit-current-thread asm (from-sp dst)))
         (($ $primcall 'cached-toplevel-box (scope name bound?))
          (emit-cached-toplevel-box asm (from-sp dst)
                                    (constant scope) (constant name)
@@ -193,8 +181,49 @@
         (($ $primcall 'struct-ref/immediate (struct n))
          (emit-struct-ref/immediate asm (from-sp dst) (from-sp (slot struct))
                                     (constant n)))
+        (($ $primcall 'add/immediate (x y))
+         (emit-add/immediate asm (from-sp dst) (from-sp (slot x)) (constant y)))
+        (($ $primcall 'sub/immediate (x y))
+         (emit-sub/immediate asm (from-sp dst) (from-sp (slot x)) (constant y)))
+        (($ $primcall 'uadd/immediate (x y))
+         (emit-uadd/immediate asm (from-sp dst) (from-sp (slot x))
+                              (constant y)))
+        (($ $primcall 'usub/immediate (x y))
+         (emit-usub/immediate asm (from-sp dst) (from-sp (slot x))
+                              (constant y)))
+        (($ $primcall 'umul/immediate (x y))
+         (emit-umul/immediate asm (from-sp dst) (from-sp (slot x))
+                              (constant y)))
+        (($ $primcall 'ursh/immediate (x y))
+         (emit-ursh/immediate asm (from-sp dst) (from-sp (slot x))
+                              (constant y)))
+        (($ $primcall 'ulsh/immediate (x y))
+         (emit-ulsh/immediate asm (from-sp dst) (from-sp (slot x))
+                              (constant y)))
         (($ $primcall 'builtin-ref (name))
          (emit-builtin-ref asm (from-sp dst) (constant name)))
+        (($ $primcall 'scm->f64 (src))
+         (emit-scm->f64 asm (from-sp dst) (from-sp (slot src))))
+        (($ $primcall 'load-f64 (src))
+         (emit-load-f64 asm (from-sp dst) (constant src)))
+        (($ $primcall 'f64->scm (src))
+         (emit-f64->scm asm (from-sp dst) (from-sp (slot src))))
+        (($ $primcall 'scm->u64 (src))
+         (emit-scm->u64 asm (from-sp dst) (from-sp (slot src))))
+        (($ $primcall 'scm->u64/truncate (src))
+         (emit-scm->u64/truncate asm (from-sp dst) (from-sp (slot src))))
+        (($ $primcall 'load-u64 (src))
+         (emit-load-u64 asm (from-sp dst) (constant src)))
+        (($ $primcall 'u64->scm (src))
+         (emit-u64->scm asm (from-sp dst) (from-sp (slot src))))
+        (($ $primcall 'scm->s64 (src))
+         (emit-scm->s64 asm (from-sp dst) (from-sp (slot src))))
+        (($ $primcall 'load-s64 (src))
+         (emit-load-s64 asm (from-sp dst) (constant src)))
+        (($ $primcall 's64->scm (src))
+         (emit-s64->scm asm (from-sp dst) (from-sp (slot src))))
+        (($ $primcall 'bv-length (bv))
+         (emit-bv-length asm (from-sp dst) (from-sp (slot bv))))
         (($ $primcall 'bv-u8-ref (bv idx))
          (emit-bv-u8-ref asm (from-sp dst) (from-sp (slot bv))
                          (from-sp (slot idx))))
@@ -376,12 +405,21 @@
         ;; the set of macro-instructions in assembly.scm.
         (($ $primcall 'eq? (a b)) (binary emit-br-if-eq a b))
         (($ $primcall 'eqv? (a b)) (binary emit-br-if-eqv a b))
-        (($ $primcall 'equal? (a b)) (binary emit-br-if-equal a b))
         (($ $primcall '< (a b)) (binary emit-br-if-< a b))
         (($ $primcall '<= (a b)) (binary emit-br-if-<= a b))
         (($ $primcall '= (a b)) (binary emit-br-if-= a b))
         (($ $primcall '>= (a b)) (binary emit-br-if-<= b a))
         (($ $primcall '> (a b)) (binary emit-br-if-< b a))
+        (($ $primcall 'u64-< (a b)) (binary emit-br-if-u64-< a b))
+        (($ $primcall 'u64-<= (a b)) (binary emit-br-if-u64-<= a b))
+        (($ $primcall 'u64-= (a b)) (binary emit-br-if-u64-= a b))
+        (($ $primcall 'u64->= (a b)) (binary emit-br-if-u64-<= b a))
+        (($ $primcall 'u64-> (a b)) (binary emit-br-if-u64-< b a))
+        (($ $primcall 'u64-<-scm (a b)) (binary emit-br-if-u64-<-scm a b))
+        (($ $primcall 'u64-<=-scm (a b)) (binary emit-br-if-u64-<=-scm a b))
+        (($ $primcall 'u64-=-scm (a b)) (binary emit-br-if-u64-=-scm a b))
+        (($ $primcall 'u64->=-scm (a b)) (binary emit-br-if-u64->=-scm a b))
+        (($ $primcall 'u64->-scm (a b)) (binary emit-br-if-u64->-scm a b))
         (($ $primcall 'logtest (a b)) (binary emit-br-if-logtest a b))))
 
     (define (compile-trunc label k exp nreq rest-var)
@@ -393,8 +431,7 @@
                      ((src . dst) (emit-mov asm (from-sp dst) (from-sp src))))
                     (lookup-parallel-moves label allocation))
           (emit-call asm proc-slot nargs)
-          (emit-dead-slot-map asm proc-slot
-                              (lookup-dead-slot-map label allocation))
+          (emit-slot-map asm proc-slot (lookup-slot-map label allocation))
           (cond
            ((and (= 1 nreq) (and rest-var) (not (maybe-slot rest-var))
                  (match (lookup-parallel-moves k allocation)
@@ -484,13 +521,16 @@
            (emit-label asm label)
            (set! frame-size (lookup-nlocals label allocation))
            (emit-begin-kw-arity asm req opt rest kw-indices allow-other-keys?
-                                frame-size alt)))
+                                frame-size alt)
+           ;; All arities define a closure binding in slot 0.
+           (emit-definition asm 'closure 0 'scm)))
         (($ $kargs names vars ($ $continue k src exp))
          (emit-label asm label)
          (for-each (lambda (name var)
                      (let ((slot (maybe-slot var)))
                        (when slot
-                         (emit-definition asm name slot))))
+                         (let ((repr (lookup-representation var allocation)))
+                           (emit-definition asm name slot repr)))))
                    names vars)
          (when src
            (emit-source asm src))
