@@ -62,6 +62,7 @@
             &module
             &struct
             &string
+            &thread
             &bytevector
             &closure
 
@@ -170,6 +171,9 @@
   ;; Indicates that an expression depends on the current module.
   &module
 
+  ;; Indicates that an expression depends on the current thread.
+  &thread
+
   ;; Indicates that an expression depends on the value of a struct
   ;; field.  The effect field indicates the specific field, or zero for
   ;; an unknown field.
@@ -285,6 +289,12 @@ is or might be a read or a write to the same location as A."
   ((push-fluid f v)                (&write-object &fluid)      &type-check)
   ((pop-fluid)                     (&write-object &fluid)      &type-check))
 
+;; Threads.  Calls cause &all-effects, which reflects the fact that any
+;; call can capture a partial continuation and reinstate it on another
+;; thread.
+(define-primitive-effects
+  ((current-thread)                (&read-object &thread)))
+
 ;; Prompts.
 (define-primitive-effects
   ((make-prompt-tag #:optional arg) (&allocate &unknown-memory-kinds)))
@@ -317,7 +327,7 @@ is or might be a read or a write to the same location as A."
   (logior &write (vector-field n constants)))
 (define-primitive-effects* constants
   ((vector . _)                    (&allocate &vector))
-  ((make-vector n init)            (&allocate &vector)         &type-check)
+  ((make-vector n init)            (&allocate &vector))
   ((make-vector/immediate n init)  (&allocate &vector))
   ((vector-ref v n)                (read-vector-field n constants) &type-check)
   ((vector-ref/immediate v n)      (read-vector-field n constants) &type-check)
@@ -351,9 +361,22 @@ is or might be a read or a write to the same location as A."
   ((string->number _)              (&read-object &string)      &type-check)
   ((string-length s)                                           &type-check))
 
+;; Unboxed floats and integers.
+(define-primitive-effects
+  ((scm->f64 _)                                                &type-check)
+  ((load-f64 _))
+  ((f64->scm _))
+  ((scm->u64 _)                                                &type-check)
+  ((scm->u64/truncate _)                                       &type-check)
+  ((load-u64 _))
+  ((u64->scm _))
+  ((scm->s64 _)                                                &type-check)
+  ((load-s64 _))
+  ((s64->scm _)))
+
 ;; Bytevectors.
 (define-primitive-effects
-  ((bytevector-length _)                                       &type-check)
+  ((bv-length _)                                               &type-check)
 
   ((bv-u8-ref bv n)                (&read-object &bytevector)  &type-check)
   ((bv-s8-ref bv n)                (&read-object &bytevector)  &type-check)
@@ -404,13 +427,33 @@ is or might be a read or a write to the same location as A."
   ((> . _)                         &type-check)
   ((<= . _)                        &type-check)
   ((>= . _)                        &type-check)
+  ((u64-= . _))
+  ((u64-< . _))
+  ((u64-> . _))
+  ((u64-<= . _))
+  ((u64->= . _))
+  ((u64-<-scm . _)                 &type-check)
+  ((u64-<=-scm . _)                &type-check)
+  ((u64-=-scm . _)                 &type-check)
+  ((u64->=-scm . _)                 &type-check)
+  ((u64->-scm . _)                 &type-check)
   ((zero? . _)                     &type-check)
   ((add . _)                       &type-check)
+  ((add/immediate . _)             &type-check)
   ((mul . _)                       &type-check)
   ((sub . _)                       &type-check)
+  ((sub/immediate . _)             &type-check)
   ((div . _)                       &type-check)
-  ((sub1 . _)                      &type-check)
-  ((add1 . _)                      &type-check)
+  ((fadd . _))
+  ((fsub . _))
+  ((fmul . _))
+  ((fdiv . _))
+  ((uadd . _))
+  ((usub . _))
+  ((umul . _))
+  ((uadd/immediate . _))
+  ((usub/immediate . _))
+  ((umul/immediate . _))
   ((quo . _)                       &type-check)
   ((rem . _)                       &type-check)
   ((mod . _)                       &type-check)
@@ -428,7 +471,15 @@ is or might be a read or a write to the same location as A."
   ((logand . _)                    &type-check)
   ((logior . _)                    &type-check)
   ((logxor . _)                    &type-check)
+  ((logsub . _)                    &type-check)
   ((lognot . _)                    &type-check)
+  ((ulogand . _))
+  ((ulogior . _))
+  ((ulogsub . _))
+  ((ursh . _))
+  ((ulsh . _))
+  ((ursh/immediate . _))
+  ((ulsh/immediate . _))
   ((logtest a b)                   &type-check)
   ((logbit? a b)                   &type-check)
   ((sqrt _)                        &type-check)
