@@ -249,7 +249,7 @@
         (with-cps cps
           (let$ body (with-cps-constants ((unspecified *unspecified*))
                        (build-term
-                         ($continue k src ($primcall 'return (unspecified))))))
+                         ($continue k src ($values (unspecified))))))
           (letk kvoid ($kargs () () ,body))
           kvoid))
        (($ $kreceive arity kargs)
@@ -287,7 +287,7 @@
         (with-cps cps
           (letv val)
           (letk kval ($kargs ('val) (val)
-                       ($continue k src ($primcall 'return (val)))))
+                       ($continue k src ($values (val)))))
           kval))
        (($ $kreceive arity kargs)
         (match arity
@@ -506,6 +506,18 @@
 
     (($ <primcall> src name args)
      (cond
+      ((eq? name 'equal?)
+       (convert-args cps args
+         (lambda (cps args)
+           (with-cps cps
+             (let$ k* (adapt-arity k src 1))
+             (letk kt ($kargs () () ($continue k* src ($const #t))))
+             (letk kf* ($kargs () ()
+                         ;; Here we continue to the original $kreceive
+                         ;; or $ktail, as equal? doesn't have a VM op.
+                         ($continue k src ($primcall 'equal? args))))
+             (build-term ($continue kf* src
+                           ($branch kt ($primcall 'eqv? args))))))))
       ((branching-primitive? name)
        (convert-args cps args
          (lambda (cps args)
@@ -555,6 +567,110 @@
                       ($ (lp args ktail)))))))))))
       ((prim-instruction name)
        => (lambda (instruction)
+            (define (box+adapt-arity cps k src out)
+              (case instruction
+                ((bv-f32-ref bv-f64-ref)
+                 (with-cps cps
+                   (letv f64)
+                   (let$ k (adapt-arity k src out))
+                   (letk kbox ($kargs ('f64) (f64)
+                                ($continue k src ($primcall 'f64->scm (f64)))))
+                   kbox))
+                ((string-length
+                  vector-length
+                  bv-length bv-u8-ref bv-u16-ref bv-u32-ref bv-u64-ref)
+                 (with-cps cps
+                   (letv u64)
+                   (let$ k (adapt-arity k src out))
+                   (letk kbox ($kargs ('u64) (u64)
+                                ($continue k src ($primcall 'u64->scm (u64)))))
+                   kbox))
+                ((bv-s8-ref bv-s16-ref bv-s32-ref bv-s64-ref)
+                 (with-cps cps
+                   (letv s64)
+                   (let$ k (adapt-arity k src out))
+                   (letk kbox ($kargs ('s64) (s64)
+                                ($continue k src ($primcall 's64->scm (s64)))))
+                   kbox))
+                (else
+                 (adapt-arity cps k src out))))
+            (define (unbox-arg cps arg unbox-op have-arg)
+              (with-cps cps
+                (letv unboxed)
+                (let$ body (have-arg unboxed))
+                (letk kunboxed ($kargs ('unboxed) (unboxed) ,body))
+                (build-term
+                  ($continue kunboxed src ($primcall unbox-op (arg))))))
+            (define (unbox-args cps args have-args)
+              (case instruction
+                ((bv-f32-ref bv-f64-ref
+                  bv-s8-ref bv-s16-ref bv-s32-ref bv-s64-ref
+                  bv-u8-ref bv-u16-ref bv-u32-ref bv-u64-ref)
+                 (match args
+                   ((bv idx)
+                    (unbox-arg
+                     cps idx 'scm->u64
+                     (lambda (cps idx)
+                       (have-args cps (list bv idx)))))))
+                ((bv-f32-set! bv-f64-set!)
+                 (match args
+                   ((bv idx val)
+                    (unbox-arg
+                     cps idx 'scm->u64
+                     (lambda (cps idx)
+                       (unbox-arg
+                        cps val 'scm->f64
+                        (lambda (cps val)
+                          (have-args cps (list bv idx val)))))))))
+                ((bv-s8-set! bv-s16-set! bv-s32-set! bv-s64-set!)
+                 (match args
+                   ((bv idx val)
+                    (unbox-arg
+                     cps idx 'scm->u64
+                     (lambda (cps idx)
+                       (unbox-arg
+                        cps val 'scm->s64
+                        (lambda (cps val)
+                          (have-args cps (list bv idx val)))))))))
+                ((bv-u8-set! bv-u16-set! bv-u32-set! bv-u64-set!)
+                 (match args
+                   ((bv idx val)
+                    (unbox-arg
+                     cps idx 'scm->u64
+                     (lambda (cps idx)
+                       (unbox-arg
+                        cps val 'scm->u64
+                        (lambda (cps val)
+                          (have-args cps (list bv idx val)))))))))
+                ((vector-ref struct-ref string-ref)
+                 (match args
+                   ((obj idx)
+                    (unbox-arg
+                     cps idx 'scm->u64
+                     (lambda (cps idx)
+                       (have-args cps (list obj idx)))))))
+                ((vector-set! struct-set!)
+                 (match args
+                   ((obj idx val)
+                    (unbox-arg
+                     cps idx 'scm->u64
+                     (lambda (cps idx)
+                       (have-args cps (list obj idx val)))))))
+                ((make-vector)
+                 (match args
+                   ((length init)
+                    (unbox-arg
+                     cps length 'scm->u64
+                     (lambda (cps length)
+                       (have-args cps (list length init)))))))
+                ((allocate-struct)
+                 (match args
+                   ((vtable nfields)
+                    (unbox-arg
+                     cps nfields 'scm->u64
+                     (lambda (cps nfields)
+                       (have-args cps (list vtable nfields)))))))
+                (else (have-args cps args))))
             (convert-args cps args
               (lambda (cps args)
                 ;; Tree-IL primcalls are sloppy, in that it could be
@@ -566,10 +682,14 @@
                   ((out . in)
                    (if (= in (length args))
                        (with-cps cps
-                         (let$ k (adapt-arity k src out))
-                         (build-term
-                           ($continue k src
-                             ($primcall instruction args))))
+                         (let$ k (box+adapt-arity k src out))
+                         ($ (unbox-args
+                             args
+                             (lambda (cps args)
+                               (with-cps cps
+                                 (build-term
+                                   ($continue k src
+                                     ($primcall instruction args))))))))
                        (with-cps cps
                          (letv prim)
                          (letk kprim ($kargs ('prim) (prim)
@@ -653,7 +773,7 @@
            (build-term ($continue k src ($primcall 'apply args*)))))))
 
     (($ <conditional> src test consequent alternate)
-     (define (convert-test cps kt kf)
+     (define (convert-test cps test kt kf)
        (match test
          (($ <primcall> src (? branching-primitive? name) args)
           (convert-args cps args
@@ -661,6 +781,13 @@
               (with-cps cps
                 (build-term ($continue kf src
                               ($branch kt ($primcall name args))))))))
+         (($ <conditional> src test consequent alternate)
+          (with-cps cps
+            (let$ t (convert-test consequent kt kf))
+            (let$ f (convert-test alternate kt kf))
+            (letk kt* ($kargs () () ,t))
+            (letk kf* ($kargs () () ,f))
+            ($ (convert-test test kt* kf*))))
          (_ (convert-arg cps test
               (lambda (cps test)
                 (with-cps cps
@@ -671,7 +798,7 @@
        (let$ f (convert alternate k subst))
        (letk kt ($kargs () () ,t))
        (letk kf ($kargs () () ,f))
-       ($ (convert-test kt kf))))
+       ($ (convert-test test kt kf))))
 
     (($ <lexical-set> src name gensym exp)
      (convert-arg cps exp
@@ -892,6 +1019,16 @@ integer."
                                                    (make-lexical-ref src 'v v)))
                               (make-lexical-ref src 'v v)))))
 
+       ;; Lower (logand x (lognot y)) to (logsub x y).  We do it here
+       ;; instead of in CPS because it gets rid of the lognot entirely;
+       ;; if type folding can't prove Y to be an exact integer, then DCE
+       ;; would have to leave it in the program for its possible
+       ;; effects.
+       (($ <primcall> src 'logand (x ($ <primcall> _ 'lognot (y))))
+        (make-primcall src 'logsub (list x y)))
+       (($ <primcall> src 'logand (($ <primcall> _ 'lognot (y)) x))
+        (make-primcall src 'logsub (list x y)))
+
        (($ <prompt> src escape-only? tag body
            ($ <lambda> hsrc hmeta
               ($ <lambda-case> _ hreq #f hrest #f () hsyms hbody #f)))
